@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/child.c,v 1.5.2.7 2007/10/12 04:32:13 y-asaba Exp $
+ * $Header: /cvsroot/pgpool/pgpool-II/child.c,v 1.16 2007/10/12 04:51:22 y-asaba Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -32,6 +32,9 @@
 #ifdef HAVE_NETINET_TCP_H
 #include <netinet/tcp.h>
 #endif
+#ifdef HAVE_SYS_SELECT_H
+#include <sys/select.h>
+#endif
 
 #include <signal.h>
 
@@ -50,11 +53,6 @@
 #include "pool_ip.h"
 #include "md5.h"
 
-#ifdef NONE_BLOCK
-static void set_nonblock(int fd);
-static void unset_nonblock(int fd);
-#endif
-
 void child_exit(int code);
 
 static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *timeout);
@@ -63,6 +61,8 @@ static POOL_CONNECTION_POOL *connect_backend(StartupPacket *sp, POOL_CONNECTION
 static void cancel_request(CancelPacket *sp, int secondary_backend);
 static RETSIGTYPE die(int sig);
 static RETSIGTYPE close_idle_connection(int sig);
+static RETSIGTYPE wakeup_handler(int sig);
+static RETSIGTYPE reload_config_handler(int sig);
 static RETSIGTYPE authentication_timeout(int sig);
 static int send_params(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
 static void send_frontend_exits(void);
@@ -74,12 +74,15 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password);
 static int exit_request;
 
 static int idle;		/* non 0 means this child is in idle state */
+static int accepted = 0;
 
 extern int myargc;
 extern char **myargv;
 
 char remote_ps_data[NI_MAXHOST];		/* used for set_ps_display */
 
+volatile sig_atomic_t got_sighup = 0;
+
 /*
 * child main loop
 */
@@ -106,19 +109,19 @@ void do_child(int unix_fd, int inet_fd)
 	signal(SIGALRM, SIG_DFL);
 	signal(SIGTERM, die);
 	signal(SIGINT, die);
-	signal(SIGHUP, close_idle_connection);
+	signal(SIGHUP, reload_config_handler);
 	signal(SIGQUIT, die);
 	signal(SIGCHLD, SIG_DFL);
-	signal(SIGUSR1, SIG_DFL);
-	signal(SIGUSR2, SIG_DFL);
+	signal(SIGUSR1, close_idle_connection);
+	signal(SIGUSR2, wakeup_handler);
 	signal(SIGPIPE, SIG_IGN);
 
 #ifdef NONE_BLOCK
 	/* set listen fds to none block */
-	set_nonblock(unix_fd);
+	pool_set_nonblock(unix_fd);
 	if (inet_fd)
 	{
-		set_nonblock(inet_fd);
+		pool_set_nonblock(inet_fd);
 	}
 #endif
 
@@ -173,6 +176,7 @@ void do_child(int unix_fd, int inet_fd)
 		}
 
 		idle = 1;
+		accepted = 0;
 
 		/* perform accept() */
 		frontend = do_accept(unix_fd, inet_fd, &timeout);
@@ -191,11 +195,7 @@ void do_child(int unix_fd, int inet_fd)
 		}
 
 		/* set frontend fd to blocking */
-		unset_nonblock(frontend->fd);
-
-#ifdef NOT_USED
-		set_nonblock(frontend->fd);
-#endif
+		pool_unset_nonblock(frontend->fd);
 
 		/* set busy flag and clear child idle timer */
 		idle = 0;
@@ -218,6 +218,7 @@ void do_child(int unix_fd, int inet_fd)
 		{
 			/* failed to read the startup packet. return to the accept() loop */
 			pool_close(frontend);
+			connection_count_down();
 			continue;
 		}
 
@@ -229,6 +230,7 @@ void do_child(int unix_fd, int inet_fd)
 				cancel_request((CancelPacket *)sp->startup_packet, 1);
 			pool_close(frontend);
 			pool_free_startup_packet(sp);
+			connection_count_down();
 			continue;
 		}
 
@@ -241,6 +243,7 @@ void do_child(int unix_fd, int inet_fd)
 			{
 				pool_close(frontend);
 				pool_free_startup_packet(sp);
+				connection_count_down();
 				continue;
 			}
 
@@ -325,7 +328,10 @@ void do_child(int unix_fd, int inet_fd)
 			connection_reuse = 0;
 
 			if ((backend = connect_backend(sp, frontend)) == NULL)
+			{
+				connection_count_down();
 				continue;
+			}
 
 			/* in master/slave mode, the first "ready for query"
 			 * packet should be treated as if we were not in the
@@ -359,6 +365,7 @@ void do_child(int unix_fd, int inet_fd)
 			if (pool_do_reauth(frontend, backend))
 			{
 				pool_close(frontend);
+				connection_count_down();
 				continue;
 			}
 
@@ -367,6 +374,7 @@ void do_child(int unix_fd, int inet_fd)
 				if (send_params(frontend, backend))
 				{
 					pool_close(frontend);
+					connection_count_down();
 					continue;
 				}
 			}
@@ -388,6 +396,7 @@ void do_child(int unix_fd, int inet_fd)
 			if (pool_flush(frontend) < 0)
 			{
 				pool_close(frontend);
+				connection_count_down();
 				continue;
 			}
 
@@ -405,6 +414,9 @@ void do_child(int unix_fd, int inet_fd)
 				 sp->user, sp->database, remote_ps_data);
 		set_ps_display(psbuf, false);
 
+		if (MAJOR(backend) == PROTO_MAJOR_V2)
+			TSTATE(backend) = 'I';
+
 		/* select load balancing node */
 		backend->info->load_balancing_node = select_load_balancing_node();
 
@@ -485,6 +497,9 @@ void do_child(int unix_fd, int inet_fd)
 				break;
 		}
 
+		accepted = 0;
+		connection_count_down();
+
 		timeout.tv_sec = pool_config->child_life_time;
 		timeout.tv_usec = 0;
 
@@ -509,11 +524,10 @@ void do_child(int unix_fd, int inet_fd)
  * -------------------------------------------------------------------
  */
 
-#ifdef NONE_BLOCK
 /*
  * set non-block flag
  */
-static void set_nonblock(int fd)
+void pool_set_nonblock(int fd)
 {
 	int var;
 
@@ -530,12 +544,11 @@ static void set_nonblock(int fd)
 		child_exit(1);
 	}
 }
-#endif
 
 /*
  * unset non-block flag
  */
-static void unset_nonblock(int fd)
+void pool_unset_nonblock(int fd)
 {
 	int var;
 
@@ -680,8 +693,17 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 #ifdef ACCEPT_PERFORMANCE
 	gettimeofday(&now1,0);
 #endif
-	afd = accept(fd, (struct sockaddr *)&saddr.addr, &saddr.salen);
 
+ retry_accept:
+
+	/* wait if recovery is started */
+	while (*InRecovery == 1)
+	{
+		pause();
+	}
+
+	afd = accept(fd, (struct sockaddr *)&saddr.addr, &saddr.salen);
+ 
 	save_errno = errno;
 	/* check backend timer is expired */
 	if (backend_timer_expired)
@@ -692,6 +714,9 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 	errno = save_errno;
 	if (afd < 0)
 	{
+		if (errno == EINTR && *InRecovery)
+			goto retry_accept;
+
 		/*
 		 * "Resource temporarily unavailable" (EAGAIN or EWOULDBLOCK)
 		 * can be silently ignored. And EINTR can be ignored.
@@ -710,6 +735,20 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 	}
 #endif
 
+	/* reload config file */
+	if (got_sighup)
+	{
+		pool_get_config(get_config_file_name(), RELOAD_CONFIG);
+		if (pool_config->enable_pool_hba)
+			load_hba(get_hba_file_name());
+		if (pool_config->parallel_mode)
+			pool_memset_system_db_info(system_db_info->info);
+		got_sighup = 0;
+	}
+
+	connection_count_up();
+	accepted = 1;
+
 	if (pool_config->parallel_mode)
 	{
 		/*
@@ -894,8 +933,6 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 	{
 		pool_error("read_startup_packet: invalid startup packet");
 		pool_free_startup_packet(sp);
-		alarm(0);
-		pool_signal(SIGALRM, SIG_IGN);
 		return NULL;
 	}
 
@@ -1167,7 +1204,7 @@ static RETSIGTYPE die(int sig)
 }
 
 /*
- * signal handler for SIGHUP
+ * signal handler for SIGUSR1
  * close all idle connections
  */
 static RETSIGTYPE close_idle_connection(int sig)
@@ -1290,6 +1327,9 @@ void pool_free_startup_packet(StartupPacket *sp)
 
 void child_exit(int code)
 {
+	if (accepted)
+		connection_count_down();
+
 	if(pool_config->parallel_mode || pool_config->enable_query_cache)
 	{
 		if (system_db_info->pgconn)
@@ -1685,6 +1725,28 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 	return -1;
 }
 
+void connection_count_up(void)
+{
+	pool_semaphore_lock(CONN_COUNTER_SEM);
+	Req_info->conn_counter++;
+	pool_semaphore_unlock(CONN_COUNTER_SEM);
+}
+
+void connection_count_down(void)
+{
+	pool_semaphore_lock(CONN_COUNTER_SEM);
+	Req_info->conn_counter--;
+	pool_semaphore_unlock(CONN_COUNTER_SEM);
+}
+
+/* 
+ * handle SIGUSR2
+ * Wakeup all process
+ */
+static RETSIGTYPE wakeup_handler(int sig)
+{
+}
+
 
 /*
  * Select load balancing node
@@ -1722,3 +1784,9 @@ int select_load_balancing_node(void)
 	pool_debug("select_load_balancing_node: selected backend id is %d", selected_slot);
 	return selected_slot;
 }
+
+/* SIGHUP handler */
+static RETSIGTYPE reload_config_handler(int sig)
+{
+	got_sighup = 1;
+}
