@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/main.c,v 1.6.2.2 2007/07/09 05:00:38 y-asaba Exp $
+ * $Header: /cvsroot/pgpool/pgpool-II/main.c,v 1.20 2007/11/09 04:37:35 y-asaba Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -22,12 +22,12 @@
 
 #include <ctype.h>
 #include <sys/types.h>
-#include <sys/time.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <sys/un.h>
 #include <netdb.h>
 #include <arpa/inet.h>
+#include <sys/time.h>
 #ifdef HAVE_SYS_SELECT_H
 #include <sys/select.h>
 #endif
@@ -50,9 +50,16 @@
 #endif
 
 #include "version.h"
+#include "parser/pool_memory.h"
+#include "parser/pool_string.h"
 
 #define CHECK_REQUEST \
 	do { \
+		if (wakeup_request) \
+		{ \
+			wakeup_children(); \
+			wakeup_request = 0; \
+		} \
 		if (failover_request) \
 		{ \
 			failover(); \
@@ -62,6 +69,11 @@
 		{ \
 			reaper(); \
 		} \
+		if (reload_config_request) \
+		{ \
+			reload_config(); \
+			reload_config_request = 0; \
+		} \
     } while (0)
 
 
@@ -76,17 +88,24 @@ static int create_inet_domain_socket(const char *hostname, const int port);
 static void myexit(int code);
 static void failover(void);
 static void reaper(void);
+static void wakeup_children(void);
+static void reload_config(void);
 static int pool_pause(struct timeval *timeout);
 static void pool_sleep(unsigned int second);
+static void kill_all_children(int sig);
 
 static RETSIGTYPE exit_handler(int sig);
 static RETSIGTYPE reap_handler(int sig);
 static RETSIGTYPE failover_handler(int sig);
+static RETSIGTYPE reload_config_handler(int sig);
 static RETSIGTYPE health_check_timer_handler(int sig);
+static RETSIGTYPE wakeup_handler(int sig);
 
 static void usage(void);
 static void stop_me(void);
 
+static int trigger_failover_command(int node, const char *command_line);
+
 static struct sockaddr_un un_addr;		/* unix domain socket path */
 static struct sockaddr_un pcp_un_addr;  /* unix domain socket path for PCP */
 
@@ -99,6 +118,8 @@ static int pcp_pid; /* pid for child process handling PCP */
 static int pcp_unix_fd; /* unix domain socket fd for PCP (not used) */
 static int pcp_inet_fd; /* inet domain socket fd for PCP */
 static char pcp_conf_file[POOLMAXPATHLEN+1]; /* path for pcp.conf */
+static char conf_file[POOLMAXPATHLEN+1];
+static char hba_file[POOLMAXPATHLEN+1];
 
 static int exiting = 0;		/* non 0 if I'm exiting */
 static int switching = 0;		/* non 0 if I'm fail overing or degenerating */
@@ -111,7 +132,7 @@ static int clear_cache = 0;		/* non 0 if clear chache option (-c) is given */
 static int not_detach = 0;		/* non 0 if non detach option (-n) is given */
 int debug = 0;	/* non 0 if debug option is given (-d) */
 
-static pid_t mypid;	/* pgpool parent process id */
+pid_t mypid;	/* pgpool parent process id */
 
 long int weight_master;	/* normalized weight of master (0-RAND_MAX range) */
 
@@ -120,8 +141,12 @@ static int stop_sig = SIGTERM;	/* stopping signal default value */
 static int health_check_timer_expired;		/* non 0 if health check timer expired */
 
 POOL_REQUEST_INFO *Req_info;		/* request info area in shared memory */
+volatile sig_atomic_t *InRecovery; /* non 0 if recovery is started */
+volatile sig_atomic_t reload_config_request = 0;
 static volatile sig_atomic_t failover_request = 0;
 static volatile sig_atomic_t sigchld_request = 0;
+static volatile sig_atomic_t wakeup_request = 0;
+
 static int pipe_fds[2]; /* for delivering signals */
 
 int my_proc_id;
@@ -135,8 +160,6 @@ char **myargv;
 int main(int argc, char **argv)
 {
 	int opt;
-	char conf_file[POOLMAXPATHLEN+1];
-	char hba_file[POOLMAXPATHLEN+1];
 	int i;
 	int pid;
 	int size;
@@ -224,7 +247,12 @@ int main(int argc, char **argv)
 		}
 	}
 
-	if (pool_get_config(conf_file))
+	mypid = getpid();
+
+	if (pool_init_config())
+		exit(1);
+
+	if (pool_get_config(conf_file, INIT_CONFIG))
 	{
 		pool_error("Unable to get configuration. Exiting...");
 		exit(1);
@@ -234,10 +262,31 @@ int main(int argc, char **argv)
 		load_hba(hba_file);
 
 	/*
-	 * if a non-switch argument remains, then it should be either "stop" or "switch"
+	 * if a non-switch argument remains, then it should be either "reload", "stop" or "switch"
 	 */
 	if (optind == (argc - 1))
 	{
+		if (!strcmp(argv[optind], "reload"))
+		{
+				pid_t pid;
+
+				pid = read_pid_file();
+				if (pid < 0)
+				{
+					pool_error("could not read pid file");
+					pool_shmem_exit(1);
+					exit(1);
+				}
+
+				if (kill(pid, SIGHUP) == -1)
+				{
+					pool_error("could not stop pid: %d. reason: %s", pid, strerror(errno));
+					pool_shmem_exit(1);
+					exit(1);
+				}
+				pool_shmem_exit(0);
+				exit(0);
+		}
 		if (!strcmp(argv[optind], "stop"))
 		{
 			stop_me();
@@ -292,8 +341,6 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
-	mypid = getpid();
-
 	/* clear cache */
 	if (clear_cache && pool_config->enable_query_cache && SYSDB_STATUS == CON_UP)
 	{
@@ -354,8 +401,17 @@ int main(int argc, char **argv)
 
 	/* initialize Req_info */
 	Req_info->kind = NODE_UP_REQUEST;
-	Req_info->node_id = -1;
+	memset(Req_info->node_id, -1, sizeof(int) * MAX_NUM_BACKENDS);
 	Req_info->master_node_id = 0;
+	Req_info->conn_counter = 0;
+
+	InRecovery = pool_shared_memory_create(sizeof(int));
+	if (InRecovery == NULL)
+	{
+		pool_error("failed to allocate InRecovery");
+		myexit(1);
+	}
+	*InRecovery = 0;
 
 	/* fork the children */
 	for (i=0;i<pool_config->num_init_children;i++)
@@ -371,7 +427,8 @@ int main(int argc, char **argv)
 	pool_signal(SIGQUIT, exit_handler);
 	pool_signal(SIGCHLD, reap_handler);
 	pool_signal(SIGUSR1, failover_handler);
-	pool_signal(SIGHUP, exit_handler);
+	pool_signal(SIGUSR2, wakeup_handler);
+	pool_signal(SIGHUP, reload_config_handler);
 
 	/* create pipe for delivering event */
 	if (pipe(pipe_fds) < 0)
@@ -447,7 +504,7 @@ int main(int argc, char **argv)
 					{
 						pool_log("set %d th backend down status", sts);
 						Req_info->kind = NODE_DOWN_REQUEST;
-						Req_info->node_id = sts;
+						Req_info->node_id[0] = sts;
 						failover();
 						/* need to distribute this info to children */
 					}
@@ -461,7 +518,7 @@ int main(int argc, char **argv)
 							/* retry count over */
 							pool_log("set %d th backend down status", sts);
 							Req_info->kind = NODE_DOWN_REQUEST;
-							Req_info->node_id = sts;
+							Req_info->node_id[0] = sts;
 							failover();
 							retrycnt = 0;
 						}
@@ -519,6 +576,7 @@ int main(int argc, char **argv)
 			}
 		}
 	}
+
 	pool_shmem_exit(0);
 }
 
@@ -528,6 +586,7 @@ static void usage(void)
 	fprintf(stderr, "  a generic connection pool/replication/load balance server for PostgreSQL\n\n");
 	fprintf(stderr, "usage: pgpool [-c][-f config_file][-F pcp_config_file][-a hba_file][-n][-d]\n");
 	fprintf(stderr, "usage: pgpool [-f config_file][-F pcp_config_file][-a hba_file] [-m {s[mart]|f[ast]|i[mmediate]}] stop\n");
+	fprintf(stderr, "usage: pgpool [-f config_file][-F pcp_config_file][-a hba_file] reload\n");
 	fprintf(stderr, "usage: pgpool -h\n");
 	fprintf(stderr, "  config_file default path: %s/%s\n",DEFAULT_CONFIGDIR, POOL_CONF_FILE_NAME);
 	fprintf(stderr, "  pcp_config_file default path: %s/%s\n", DEFAULT_CONFIGDIR, PCP_PASSWD_FILE_NAME);
@@ -570,6 +629,8 @@ static void daemonize(void)
 	}
 #endif
 
+	mypid = getpid();
+
 	i = open("/dev/null", O_RDWR);
 	dup2(i, 0);
 	dup2(i, 1);
@@ -685,6 +746,7 @@ pid_t pcp_fork_a_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 
 		/* call PCP child main */
 		POOL_SETMASK(&UnBlockSig);
+		reload_config_request = 0;
 		pcp_do_child(unix_fd, inet_fd, pcp_conf_file);
 	}
 	else if (pid == -1)
@@ -713,6 +775,7 @@ pid_t fork_a_child(int unix_fd, int inet_fd, int id)
 
 		/* call child main */
 		POOL_SETMASK(&UnBlockSig);
+		reload_config_request = 0;
 		my_proc_id = id;
 		do_child(unix_fd, inet_fd);
 	}
@@ -863,28 +926,40 @@ static void myexit(int code)
 	exit(code);
 }
 
-/* notice backend connection error using SIGUSR1 */
 void notice_backend_error(int node_id)
+{
+	int n = node_id;
+
+	degenerate_backend_set(&n, 1);
+}
+
+/* notice backend connection error using SIGUSR1 */
+void degenerate_backend_set(int *node_id_set, int count)
 {
 	pid_t parent = getppid();
+	int i;
 
 	if (pool_config->parallel_mode)
 	{
 		return;
 	}
 
-	pool_log("notice_backend_error: %d fail over request from pid %d", node_id, getpid());
-
+	pool_semaphore_lock(REQUEST_INFO_SEM);
 	Req_info->kind = NODE_DOWN_REQUEST;
-	Req_info->node_id = node_id;
-
-	if (node_id < 0 || node_id >= MAX_NUM_BACKENDS || !VALID_BACKEND(node_id))
+	for (i = 0; i < count; i++)
 	{
-		pool_error("notice_backend_error: node %d is not valid backend.");
-		return;
-	}
+		if (node_id_set[i] < 0 || node_id_set[i] >= MAX_NUM_BACKENDS ||
+			!VALID_BACKEND(node_id_set[i]))
+		{
+			pool_log("notice_backend_error: node %d is not valid backend.");
+			continue;
+		}
 
+		pool_log("notice_backend_error: %d fail over request from pid %d", node_id_set[i], getpid());
+		Req_info->node_id[i] = node_id_set[i];
+	}
 	kill(parent, SIGUSR1);
+	pool_semaphore_unlock(REQUEST_INFO_SEM);
 }
 
 /* send failback request using SIGUSR1 */
@@ -894,7 +969,7 @@ void send_failback_request(int node_id)
 
 	pool_log("send_failback_request: fail back %d th node request from pid %d", node_id, getpid());
 	Req_info->kind = NODE_UP_REQUEST;
-	Req_info->node_id = node_id;
+	Req_info->node_id[0] = node_id;
 
 	if (node_id < 0 || node_id >= MAX_NUM_BACKENDS || VALID_BACKEND(node_id))
 	{
@@ -929,8 +1004,6 @@ static RETSIGTYPE exit_handler(int sig)
 		pool_log("received fast shutdown request");
 	else if (sig == SIGQUIT)
 		pool_log("received immediate shutdown request");
-	else if (sig == SIGHUP)
-		pool_log("received idle connection close request");
 	else
 	{
 		pool_error("exit_handler: unknown signal received %d", sig);
@@ -951,13 +1024,6 @@ static RETSIGTYPE exit_handler(int sig)
 
 	kill(pcp_pid, sig);
 
-	if (sig == SIGHUP)
-	{
-		exiting = 0;
-		POOL_SETMASK(&UnBlockSig);
-		return;
-	}
-
 	POOL_SETMASK(&UnBlockSig);
 
 	while (wait(NULL) > 0)
@@ -989,7 +1055,6 @@ static RETSIGTYPE failover_handler(int sig)
  */
 static void failover(void)
 {
-
 	int node_id;
 	int i;
 
@@ -1002,6 +1067,7 @@ static void failover(void)
 	if (getpid() != mypid)
 	{
 		pool_debug("failover_handler: I am not parent");
+		kill(pcp_pid, SIGUSR2);
 		return;
 	}
 
@@ -1011,6 +1077,7 @@ static void failover(void)
 	if (exiting)
 	{
 		pool_debug("failover_handler called while exiting");
+		kill(pcp_pid, SIGUSR2);
 		return;
 	}
 
@@ -1020,54 +1087,75 @@ static void failover(void)
 	if (switching)
 	{
 		pool_debug("failover_handler called while switching");
+		kill(pcp_pid, SIGUSR2);
 		return;
 	}
 
-	node_id = Req_info->node_id;
+	pool_semaphore_lock(REQUEST_INFO_SEM);
 
-	if (node_id < 0)
+	if (Req_info->kind == CLOSE_IDLE_REQUEST)
 	{
-		pool_error("failover_handler: invalid node_id %d", node_id);
+		pool_semaphore_unlock(REQUEST_INFO_SEM);
+		kill_all_children(SIGUSR1);
+		kill(pcp_pid, SIGUSR2);
 		return;
 	}
 
-	if (node_id >= MAX_NUM_BACKENDS ||
-		(Req_info->kind == NODE_UP_REQUEST && VALID_BACKEND(node_id)) ||
-		(Req_info->kind == NODE_DOWN_REQUEST && !VALID_BACKEND(node_id)))
-	{
-		pool_error("failover_handler: invalid node_id %d status:%d MAX_NUM_BACKENDS: %d", node_id,
-					BACKEND_INFO(node_id).backend_status, MAX_NUM_BACKENDS);
-		return;
-	}
-
-	if (Req_info->kind == NODE_UP_REQUEST)
-	{
-		pool_log("starting fail back. reconnect host %s(%d)",
-				 BACKEND_INFO(node_id).backend_hostname,
-				 BACKEND_INFO(node_id).backend_port);
-	}
-	else
-	{
-		pool_log("starting degeneration. shutdown host %s(%d)",
-				 BACKEND_INFO(node_id).backend_hostname,
-				 BACKEND_INFO(node_id).backend_port);
-	}
-
 	/* 
 	 * if not in replication mode/master slave mode, we treat this a restart request.
 	 * otherwise we need to check if we have already failovered.
 	 */
-	pool_debug("failover_handler: starting to select new master node");
+	pool_debug("failover_handler: starting to selec new master node");
 	switching = 1;
+	node_id = Req_info->node_id[0];
 
 	/* failback request? */
 	if (Req_info->kind == NODE_UP_REQUEST)
 	{
+		if (node_id >= MAX_NUM_BACKENDS ||
+			(Req_info->kind == NODE_UP_REQUEST && VALID_BACKEND(node_id)) ||
+			(Req_info->kind == NODE_DOWN_REQUEST && !VALID_BACKEND(node_id)))
+		{
+			pool_semaphore_unlock(REQUEST_INFO_SEM);
+			pool_error("failover_handler: invalid node_id %d status:%d MAX_NUM_BACKENDS: %d", node_id,
+					   BACKEND_INFO(node_id).backend_status, MAX_NUM_BACKENDS);
+			kill(pcp_pid, SIGUSR2);
+			return;
+		}
+
+		pool_log("starting fail back. reconnect host %s(%d)",
+				 BACKEND_INFO(node_id).backend_hostname,
+				 BACKEND_INFO(node_id).backend_port);
 		BACKEND_INFO(node_id).backend_status = CON_CONNECT_WAIT;	/* unset down status */
+		trigger_failover_command(node_id, pool_config->failback_command);
 	}
 	else
 	{
-		BACKEND_INFO(node_id).backend_status = CON_DOWN;	/* set down status */
+		int cnt = 0;
+
+		for (i = 0; i < MAX_NUM_BACKENDS; i++)
+		{
+			if (Req_info->node_id[i] != -1 &&
+				VALID_BACKEND(Req_info->node_id[i]))
+			{
+				pool_log("starting degeneration. shutdown host %s(%d)",
+						 BACKEND_INFO(Req_info->node_id[i]).backend_hostname,
+						 BACKEND_INFO(Req_info->node_id[i]).backend_port);
+
+
+				BACKEND_INFO(Req_info->node_id[i]).backend_status = CON_DOWN;	/* set down status */
+				trigger_failover_command(Req_info->node_id[i], pool_config->failover_command);
+				cnt++;
+			}
+		}
+
+		if (cnt == 0)
+		{
+			pool_log("failover: no backends are degenerated");
+			pool_semaphore_unlock(REQUEST_INFO_SEM);
+			kill(pcp_pid, SIGUSR2);
+			return;
+		}
 	}
 
 	for (i=0;i<pool_config->backend_desc->num_backends;i++)
@@ -1082,7 +1170,7 @@ static void failover(void)
 	}
 	else
 	{
-		if (Req_info->master_node_id == i)
+		if (Req_info->master_node_id == i && *InRecovery == 0)
 		{
 			pool_log("failover_handler: do not restart pgpool. same master node %d was selected", i);
 			if (Req_info->kind == NODE_UP_REQUEST)
@@ -1098,7 +1186,9 @@ static void failover(void)
 						 BACKEND_INFO(node_id).backend_port);
 			}
 
+			pool_semaphore_unlock(REQUEST_INFO_SEM);
 			switching = 0;
+			kill(pcp_pid, SIGUSR2);
 			return;
 		}
 
@@ -1126,6 +1216,9 @@ static void failover(void)
 		pool_error("failover_handler: wait() failed. reason:%s", strerror(errno));
 #endif
 
+	memset(Req_info->node_id, -1, sizeof(int) * MAX_NUM_BACKENDS);
+	pool_semaphore_unlock(REQUEST_INFO_SEM);
+
 	/* fork the children */
 	for (i=0;i<pool_config->num_init_children;i++)
 	{
@@ -1146,8 +1239,8 @@ static void failover(void)
 				 BACKEND_INFO(node_id).backend_port);
 	}
 
-	Req_info->node_id = -1;
 	switching = 0;
+	kill(pcp_pid, SIGUSR2);
 }
 
 /*
@@ -1178,6 +1271,9 @@ int health_check(void)
 	char kind;
 	int i;
 
+	if (*InRecovery)
+		return 0;
+
 	memset(&mysp, 0, sizeof(mysp));
 	mysp.len = htonl(296);
 	mysp.sp.protoVersion = htonl(PROTO_MAJOR_V2 << 16);
@@ -1305,7 +1401,6 @@ system_db_health_check(void)
 	}
 
 	close(fd);
-
 	return 0;
 }
 
@@ -1449,6 +1544,63 @@ pool_get_system_db_info(void)
 	return system_db_info->info;
 }
 
+
+/* 
+ * handle SIGUSR2
+ * Wakeup all processes
+ */
+static void wakeup_children(void)
+{
+	kill_all_children(SIGUSR2);
+}
+
+
+static RETSIGTYPE wakeup_handler(int sig)
+{
+	POOL_SETMASK(&BlockSig);	
+	wakeup_request = 1;
+	write(pipe_fds[1], "\0", 1);
+	POOL_SETMASK(&UnBlockSig);	
+}
+
+/*
+ * handle SIGHUP
+ *
+ */
+static RETSIGTYPE reload_config_handler(int sig)
+{
+	POOL_SETMASK(&BlockSig);
+	reload_config_request = 1;
+	write(pipe_fds[1], "\0", 1);
+	POOL_SETMASK(&UnBlockSig);
+}
+
+static void reload_config(void)
+{
+	pool_log("reload config files.");
+	pool_get_config(conf_file, RELOAD_CONFIG);
+	if (pool_config->enable_pool_hba)
+		load_hba(hba_file);
+	if (pool_config->parallel_mode)
+		pool_memset_system_db_info(system_db_info->info);
+	kill_all_children(SIGHUP);
+}
+
+static void kill_all_children(int sig)
+{
+	int i;
+
+	/* kill all children */
+	for (i = 0; i < pool_config->num_init_children; i++)
+	{
+		pid_t pid = pids[i].pid;
+		if (pid)
+		{
+			kill(pid, sig);
+		}
+	}
+}
+
 /*
  * pool_pause: A process pauses by select().
  */
@@ -1502,3 +1654,106 @@ static void pool_sleep(unsigned int second)
 	}
 	POOL_SETMASK(&BlockSig);
 }
+
+/*
+ * get_config_file_name: return full path of pgpool.conf.
+ */
+char *get_config_file_name(void)
+{
+	return conf_file;
+}
+
+/*
+ * get_config_file_name: return full path of pool_hba.conf.
+ */
+char *get_hba_file_name(void)
+{
+	return hba_file;
+}
+
+/*
+ * trigger_failover_command: execute specified command at failover.
+ *                           command_line is null-terminated string.
+ */
+static int trigger_failover_command(int node, const char *command_line)
+{
+	int r = 0;
+	String *exec_cmd;
+	char port_buf[6];
+	char buf[2];
+	BackendInfo *info;
+
+	if (command_line == NULL || (strlen(command_line) == 0))
+		return 0;
+
+	/* check nodeID */
+	if (node < 0 || node > NUM_BACKENDS)
+		return -1;
+
+	info = pool_get_node_info(node);
+	if (!info)
+		return -1;
+
+	buf[1] = '\0';
+	pool_memory = pool_memory_create();
+	if (!pool_memory)
+	{
+		pool_error("trigger_failover_command: pool_memory_create() failed");
+		return -1;
+	}
+	exec_cmd = init_string("");
+	
+	while (*command_line)
+	{
+		if (*command_line == '%')
+		{
+			if (*(command_line + 1))
+			{
+				char val = *(command_line + 1);
+				switch (val)
+				{
+					case 'p': /* port */
+						snprintf(port_buf, sizeof(port_buf), "%d", info->backend_port);
+						string_append_char(exec_cmd, port_buf);
+						break;
+
+					case 'D': /* database directory */
+						string_append_char(exec_cmd, info->backend_data_directory);
+						break;
+
+					case 'd': /* node id */
+						snprintf(port_buf, sizeof(port_buf), "%d", node);
+						string_append_char(exec_cmd, port_buf);
+						break;
+
+					case 'h': /* host name */
+						string_append_char(exec_cmd, info->backend_hostname);
+						break;
+
+					case '%': /* escape */
+						string_append_char(exec_cmd, "%");
+						break;
+
+					default: /* ignore */
+						break;
+				}
+				command_line++;
+			}
+		} else {
+			buf[0] = *command_line;
+			string_append_char(exec_cmd, buf);
+		}
+		command_line++;
+	}
+
+	if (strlen(exec_cmd->data) != 0)
+	{
+		pool_log("execute command: %s", exec_cmd->data);
+		r = system(exec_cmd->data);
+	}
+
+	pool_memory_delete(pool_memory, 0);
+	pool_memory = NULL;
+
+	return r;
+}
