@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/parser/pool_memory.c,v 1.6 2007/05/24 06:38:30 y-asaba Exp $
+ * $Header: /cvsroot/pgpool/pgpool-II/parser/pool_memory.c,v 1.8 2007/10/15 06:11:13 y-asaba Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -32,7 +32,7 @@
 #define ALIGN 3
 #define POOL_HEADER_SIZE (sizeof (POOL_CHUNK_HEADER))
 
-POOL_MEMORY_POOL *pool_memory;
+POOL_MEMORY_POOL *pool_memory = NULL;
 
 static int get_free_index(unsigned int size);
 
@@ -248,11 +248,17 @@ POOL_MEMORY_POOL *pool_memory_create(void)
  * pool_memory_delete:
  *     Frees all memory which is allocated in the memory pool.
  */
-void pool_memory_delete(POOL_MEMORY_POOL *pool_memory)
+void pool_memory_delete(POOL_MEMORY_POOL *pool_memory, int reuse)
 {
 	POOL_BLOCK *block, *ptr;
 
-	for (block = pool_memory->blocks; block;)
+	/* Reuse the first memory block */
+	if (reuse && pool_memory->blocks)
+		block = pool_memory->blocks->next;
+	else
+		block = pool_memory->blocks;
+
+	while (block)
 	{
 		ptr = block->next;
 		free(block->block);
@@ -268,7 +274,27 @@ void pool_memory_delete(POOL_MEMORY_POOL *pool_memory)
 		block = ptr;
 	}
 
-	free(pool_memory);
+	if (reuse)
+	{
+		int i;
+
+		if (pool_memory->blocks)
+		{
+			pool_memory->blocks->next = NULL;
+			pool_memory->blocks->allocsize = 0;
+			pool_memory->blocks->freepoint = pool_memory->blocks->block;
+		}
+		pool_memory->largeblocks = NULL;
+		for (i = 0; i < SLOT_NUM; i++)
+		{
+			pool_memory->freelist[i] = NULL;
+		}
+	}
+	else
+	{
+		free(pool_memory);
+		pool_memory = NULL;
+	}
 }
 
 /*
