@@ -2,9 +2,13 @@
 #define POOL_PARSER_H
 
 #include "../pool_type.h"
+#include <setjmp.h>
+
+extern jmp_buf jmpbuffer;
 
 /* integer */
 typedef signed int int32;
+typedef signed short int16;
 typedef unsigned int PoolOid;
 typedef unsigned int Index;
 typedef short AttrNumber;
