@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/pool_auth.c,v 1.3 2007/01/04 17:27:10 devrim Exp $
+ * $Header: /cvsroot/pgpool/pgpool-II/pool_auth.c,v 1.7 2007/11/16 02:52:51 y-asaba Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -36,6 +36,8 @@
 #include <errno.h>
 #include <string.h>
 
+#define AUTHFAIL_ERRORCODE "28000"
+
 static POOL_STATUS pool_send_auth_ok(POOL_CONNECTION *frontend, int pid, int key, int protoMajor);
 static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth, int protoMajor);
 static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth, int protoMajor);
@@ -51,6 +53,7 @@ int pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 	int key;
 	int protoMajor;
 	int length;
+	int authkind;
 	int i;
 	StartupPacket *sp;
 
@@ -103,17 +106,17 @@ int pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 	 * in non replication mode, we support kind = 0, 3, 4, 5
 	 */
 
-	pid = pool_read_int(cp);
-	if (pid < 0)
+	authkind = pool_read_int(cp);
+	if (authkind < 0)
 	{
-		pool_error("pool_do_auth: read pid failed");
+		pool_error("pool_do_auth: read auth kind failed");
 		return -1;
 	}
 
-	pid = ntohl(pid);
+	authkind = ntohl(authkind);
 
 	/* trust? */
-	if (pid == 0)
+	if (authkind == 0)
 	{
 		int msglen;
 
@@ -134,15 +137,18 @@ int pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 	}
 
 	/* clear text password authentication? */
-	else if (pid == 3)
+	else if (authkind == 3)
 	{
 		for (i=0;i<NUM_BACKENDS;i++)
 		{
+			if (!VALID_BACKEND(i))
+				continue;
+
 			pool_debug("trying clear text password authentication");
 
-			pid = do_clear_text_password(CONNECTION(cp, i), frontend, 0, protoMajor);
+			authkind = do_clear_text_password(CONNECTION(cp, i), frontend, 0, protoMajor);
 
-			if (pid < 0)
+			if (authkind < 0)
 			{
 				pool_error("do_clear_text_password failed in slot %d", i);
 				return -1;
@@ -151,15 +157,18 @@ int pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 	}
 
 	/* crypt authentication? */
-	else if (pid == 4)
+	else if (authkind == 4)
 	{
 		for (i=0;i<NUM_BACKENDS;i++)
 		{
+			if (!VALID_BACKEND(i))
+				continue;
+
 			pool_debug("trying crypt authentication");
 
-			pid = do_crypt(CONNECTION(cp, i), frontend, 0, protoMajor);
+			authkind = do_crypt(CONNECTION(cp, i), frontend, 0, protoMajor);
 
-			if (pid < 0)
+			if (authkind < 0)
 			{
 				pool_error("do_crypt_text_password failed in slot %d", i);
 				return -1;
@@ -168,15 +177,28 @@ int pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 	}
 
 	/* md5 authentication? */
-	else if (pid == 5)
+	else if (authkind == 5)
 	{
+		if (NUM_BACKENDS > 1)
+		{
+			pool_send_error_message(frontend, protoMajor, AUTHFAIL_ERRORCODE,
+									"MD5 authentication is unsupported in replication, master-slave and parallel modes.",
+									"",
+									"check pg_hba.conf",
+									__FILE__, __LINE__);
+			return -1;
+		}
+
 		for (i=0;i<NUM_BACKENDS;i++)
 		{
+			if (!VALID_BACKEND(i))
+				continue;
+
 			pool_debug("trying md5 authentication");
 
-			pid = do_md5(CONNECTION(cp, i), frontend, 0, protoMajor);
+			authkind = do_md5(CONNECTION(cp, i), frontend, 0, protoMajor);
 
-			if (pid < 0)
+			if (authkind < 0)
 			{
 				pool_error("do_md5failed in slot %d", i);
 				return -1;
@@ -184,7 +206,13 @@ int pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 		}
 	}
 
-	if (pid != 0)
+	else
+	{
+		pool_error("pool_do_auth: unsupported auth kind received: %d", authkind);
+		return -1;
+	}
+
+	if (authkind != 0)
 	{
 		pool_error("pool_do_auth: backend does not return authentication ok");
 		return -1;
