@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
-* $Header: /cvsroot/pgpool/pgpool-II/pool_stream.c,v 1.2.2.1 2007/07/11 02:15:30 y-asaba Exp $
+* $Header: /cvsroot/pgpool/pgpool-II/pool_stream.c,v 1.6 2007/11/14 03:59:58 y-mori Exp $
 *
 * pgpool: a language independent connection pool server for PostgreSQL 
 * written by Tatsuo Ishii
@@ -24,6 +24,10 @@
 
 #include "config.h"
 
+#ifdef HAVE_SYS_SELECT_H
+#include <sys/select.h>
+#endif
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -151,9 +155,107 @@ int pool_read(POOL_CONNECTION *cp, void *buf, int len)
 
 			if (cp->isbackend)
 			{
+			    /* fatal error, notice to parent and exit */
+				notice_backend_error(cp->db_node_id);
+				child_exit(1);
+			}
+			else
+			{
+			    return -1;
+			}
+		}
+		else if (readlen == 0)
+		{
+			if (cp->isbackend)
+			{
+				pool_error("pool_read2: EOF encountered with backend");
+				return -1;
+
+#ifdef NOT_USED
 			    /* fatal error, notice to parent and exit */
 			    notice_backend_error(IS_MASTER_NODE_ID(cp->db_node_id));
-			    child_exit(1);
+				child_exit(1);
+#endif
+			}
+			else
+			{
+				/*
+				 * if backend offers authentication method, frontend could close connection
+				 */
+				return -1;
+			}
+		}
+
+		if (len < readlen)
+		{
+			/* overrun. we need to save remaining data to pending buffer */
+			if (save_pending_data(cp, readbuf+len, readlen-len))
+				return -1;
+			memmove(buf, readbuf, len);
+			break;
+		}
+
+		memmove(buf, readbuf, readlen);
+		buf += readlen;
+		len -= readlen;
+	}
+
+	return 0;
+}
+
+/*
+* this function is used by SELECT QUERY at parallel query
+* read len bytes from cp
+* returns 0 on success otherwise -1.
+*/
+int pool_read_parallel(POOL_CONNECTION *cp, void *buf, int len)
+{
+	static char readbuf[READBUFSZ];
+
+	int consume_size;
+	int readlen;
+	int notimeout;
+
+	consume_size = consume_pending_data(cp, buf, len);
+	len -= consume_size;
+	buf += consume_size;
+
+	/* if this is not the master backend, then set timeout */
+	notimeout = 1;
+
+	while (len > 0)
+	{
+		if (pool_check_fd(cp, notimeout))
+		{
+			if (!IS_MASTER_NODE_ID(cp->db_node_id))
+			{
+				pool_log("pool_read: data is not ready in DB node: %d. abort this session",
+						 cp->db_node_id);
+				exit(1);
+			}
+			else
+			{
+				pool_error("pool_read: pool_check_fd failed (%s)", strerror(errno));
+			    return -1;
+			}
+		}
+
+		readlen = read(cp->fd, readbuf, READBUFSZ);
+		if (readlen == -1)
+		{
+			if (errno == EINTR || errno == EAGAIN)
+			{
+				pool_debug("pool_read: retrying due to %s", strerror(errno));
+				continue;
+			}
+
+			pool_error("pool_read: read failed (%s)", strerror(errno));
+
+			if (cp->isbackend)
+			{
+			    /* fatal error, notice to parent and exit */
+				notice_backend_error(cp->db_node_id);
+				child_exit(1);
 			}
 			else
 			{
@@ -266,8 +368,8 @@ char *pool_read2(POOL_CONNECTION *cp, int len)
 			if (cp->isbackend)
 			{
 			    /* fatal error, notice to parent and exit */
-			    notice_backend_error(IS_MASTER_NODE_ID(cp->db_node_id));
-			    child_exit(1);
+				notice_backend_error(cp->db_node_id);
+				child_exit(1);
 			}
 			else
 			{
@@ -449,7 +551,7 @@ int pool_flush(POOL_CONNECTION *cp)
 	{
 		if (cp->isbackend)
 		{
-			notice_backend_error(IS_MASTER_NODE_ID(cp->db_node_id));
+			notice_backend_error(cp->db_node_id);
 			child_exit(1);
 		}
 		else
@@ -576,8 +678,8 @@ char *pool_read_string(POOL_CONNECTION *cp, int *len, int line)
 
 			if (cp->isbackend)
 			{
-			    notice_backend_error(IS_MASTER_NODE_ID(cp->db_node_id));
-			    child_exit(1);
+				notice_backend_error(cp->db_node_id);
+				child_exit(1);
 			}
 			else
 			{
